local modkeys = {}

modkeys.map = {
  ["left command"]   = "cmd",
  ["right command"]  = "cmd",
  ["left ctrl"]      = "ctrl",
  ["right ctrl"]     = "ctrl",
  ["left shift"]     = "shift",
  ["right shift"]    = "shift",
  ["left option"]    = "option",
  ["right option"]   = "option",
  ["left alt"]       = "alt",
  ["right alt"]      = "altgr",
}

modkeys.keys = { "ctrl", "alt", "option", "altgr", "shift", "cmd" }

return modkeys
