local modkeys = {}

modkeys.map = {
  ["left ctrl"]   = "ctrl",
  ["right ctrl"]  = "ctrl",
  ["left shift"]  = "shift",
  ["right shift"] = "shift",
  ["left alt"]    = "alt",
  ["right alt"]   = "altgr",
}

modkeys.keys = { "ctrl", "shift", "alt", "altgr" }

return modkeys
